from typing import Any, Dict, Optional

from cloudinary.models import CloudinaryField
from django.db import models
from django.utils.translation import gettext_lazy as _

from ...conf import settings
from ...utils import filesizeformat
from .base import CloudinaryFieldFile, CloudinaryFileResource


class CloudinaryMedia(CloudinaryFileResource):
    file = CloudinaryField(
        _("file"),
        type=settings.CLOUDINARY_TYPE,
        resource_type="video",
    )
    display_name = models.CharField(_("display name"), max_length=255, blank=True)

    class Meta(CloudinaryFileResource.Meta):
        verbose_name = _("media")
        verbose_name_plural = _("media")

    def get_file_folder(self) -> str:
        return settings.FILES_UPLOAD_TO

    def get_file(self) -> Optional[CloudinaryFieldFile]:
        if not self.file:
            return None
        return CloudinaryFieldFile(self.file, checksum=self.checksum)

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> CloudinaryField:
        return self._meta.get_field("file")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.basename
        super().save(*args, **kwargs)

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "name": self.display_name,
            "file_info": "({ext}, {size})".format(
                ext=self.extension, size=filesizeformat(self.size)
            ),
        }

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "acceptFiles": [
                "audio/*",
                "video/*"
            ],
        }



import os
from typing import Any, Dict, Iterable, Optional, Tuple

from django.core.files import File
from django.db import models
from django.db.models.fields.files import FieldFile
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from PIL import Image
from variations.typing import Size
from variations.utils import prepare_image, replace_extension

from .. import exceptions, helpers, signals, utils
from ..conf import settings
from ..files import VariationFile
from ..typing import FileLike
from ..variations import PaperVariation
from .mixins import BacklinkModelMixin, FileFieldProxyMixin, FileProxyMixin

__all__ = [
    "NoPermissionsMetaBase",
    "ResourceBaseMeta",
    "Resource",
    "FileResource",
    "FileFieldResource",
    "ImageFileResourceMixin",
    "VariationFile",
    "VersatileImageResourceMixin",
]


class Permissions(models.Model):
    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("upload", "Can upload files"),
            ("change", "Can change files"),
            ("delete", "Can delete files"),
        )


class NoPermissionsMetaBase:
    """
    Отменяет создание автоматических объектов Permission у наследников класса.
    """
    def __new__(mcs, name, bases, attrs, **kwargs):
        meta = attrs.pop("Meta", None)
        if meta is None:
            meta = type("Meta", (), {"default_permissions": ()})
        else:
            meta_attrs = meta.__dict__.copy()
            meta_attrs.setdefault("default_permissions", ())
            meta = type("Meta", meta.__bases__, meta_attrs)
        attrs["Meta"] = meta

        return super().__new__(mcs, name, bases, attrs, **kwargs)


class ResourceBaseMeta(NoPermissionsMetaBase, models.base.ModelBase):
    pass


class ResourceBase(models.Model, metaclass=ResourceBaseMeta):
    """
    Базовый класс ресурса.
    """
    created_at = models.DateTimeField(_("created at"), default=now, editable=False)
    modified_at = models.DateTimeField(_("changed at"), auto_now=True, editable=False)

    class Meta:
        abstract = True

    def __repr__(self):
        return "{} #{}".format(type(self).__name__, self.pk)

    def as_dict(self) -> Dict[str, Any]:
        """
        Представление объекта в виде словаря, конвертируемого в JSON.
        Служит для формирования виджета файла без перезагрузки страницы.
        """
        return {
            "id": self.pk,
            "created": self.created_at.isoformat(),
            "modified": self.modified_at.isoformat(),
        }


class Resource(BacklinkModelMixin, ResourceBase):
    """
    Ресурс.
    Включает поля, позволяющие обратиться к модели и полю,
    через которые был создан ресурс.
    """

    class Meta(ResourceBase.Meta):
        abstract = True


class FileResource(FileProxyMixin, Resource):
    """
    Подкласс ресурса, представляющего файл.
    """

    basename = models.CharField(
        _("basename"),
        max_length=255,
        editable=False,
        help_text=_("Human-readable resource name"),
    )
    extension = models.CharField(
        _("extension"),
        max_length=32,
        editable=False,
        help_text=_("Lowercase, without leading dot"),
    )
    size = models.PositiveIntegerField(_("size"), default=0, editable=False)
    checksum = models.CharField(
        _("checksum"),
        max_length=64,
        editable=False,
    )
    uploaded_at = models.DateTimeField(_("uploaded at"), default=now, editable=False)

    class Meta(Resource.Meta):
        abstract = True

    def __str__(self):
        return self.get_basename()

    def __repr__(self):
        return "{}('{}')".format(type(self).__name__, self.get_basename())

    @property
    def name(self) -> str:
        """
        Полное имя загруженного файла.
        В отличие от `self.basename` это имя включает относительный путь,
        суффикс и расширение.
        """
        raise NotImplementedError

    def get_basename(self) -> str:
        """
        Человекопонятное имя файла.
        Не содержит суффикса, которое может быть добавлено файловым хранилищем.
        """
        if self.extension:
            return "{}.{}".format(self.basename, self.extension)
        return self.basename

    def _require_file(self):
        """
        Требование наличия непустой ссылки на файл.
        Физическое существование файла не гарантируется и должно проверяться
        методом `file_exists`.
        """
        file = self.get_file()
        if not file:
            file_field = self.get_file_field()
            raise ValueError(_("The '%s' attribute has no file associated with it.") % file_field.name)

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "name": self.basename,
            "extension": self.extension,
            "size": self.size,
            "url": self.get_file_url(),
            "uploaded": self.uploaded_at.isoformat()
        }

    def update_checksum(self, file: FileLike = None) -> bool:
        file = file or self.get_file()
        old_checksum = self.checksum
        new_checksum = utils.checksum(file)
        if new_checksum and new_checksum != old_checksum:
            signals.checksum_update.send(
                sender=type(self),
                instance=self,
                checksum=new_checksum
            )
            self.checksum = new_checksum
        return old_checksum != new_checksum

    def get_file(self) -> File:
        raise NotImplementedError

    def set_file(self, value):
        """
        Обновление файла в текущей модели при переименовании / удалении.
        Применяется в Cloudinary, т.к. файловые storage устанавливают
        значения поля самостоятельно.
        """
        raise NotImplementedError

    def get_file_field(self):
        """
        Получение файлового поля модели.
        """
        raise NotImplementedError

    def get_file_size(self) -> int:
        """
        Получение размера загруженного файла.
        """
        raise NotImplementedError

    def get_file_url(self) -> str:
        """
        Получение ссылки на загруженный файл.
        """
        raise NotImplementedError

    def file_exists(self) -> bool:
        """
        Проверка существования файла.
        """
        raise NotImplementedError

    def attach_file(self, file: FileLike, name: str = None, **options):
        """
        Присоединение файла к экземпляру ресурса.
        В действительности, сохранение файла происходит в методе `_attach_file`.
        Не переопределяйте этот метод, если не уверены в том, что вы делаете.

        Если на данном этапе обнаруживается, что переданный файл не может
        быть представлен этой моделью, необходимо вызвать исключение
        UnsupportedFileError.
        """
        if not isinstance(file, File):
            name = name or getattr(file, "name", None)
            if name:
                name = os.path.basename(name)
            file = File(file, name=name)
        elif name:
            file.name = name

        prepared_file = self._prepare_file(file, **options)

        signals.pre_attach_file.send(
            sender=type(self), instance=self, file=prepared_file, options=options
        )

        # reset file position before upload
        if prepared_file.seekable():
            prepared_file.seek(0)

        response = self._attach_file(prepared_file, **options)

        self.size = self.get_file_size()
        self.uploaded_at = now()
        self.modified_at = now()

        # Рассчет хэша от входного файла, а не от загруженного,
        # т.к. в случае Cloudinary, это приведет к избыточному
        # скачиванию файла из облачного хранилища.
        if prepared_file.seekable():
            prepared_file.seek(0)
            self.update_checksum(prepared_file)
        else:
            self.update_checksum()

        signals.post_attach_file.send(
            sender=type(self),
            instance=self,
            file=self.get_file(),
            options=options,
            response=response,
        )

    def _prepare_file(self, file: File, **options) -> File:
        """
        Подготовка файла к присоединению к модели.
        """
        return file

    def _attach_file(self, file: File, **options):
        raise NotImplementedError

    def rename_file(self, new_name: str, **options):
        """
        Переименование файла.
        В действительности, переименование файла происходит в методе `_rename_file`.
        Не переопределяйте этот метод, если не уверены в том, что вы делаете.
        """
        self._require_file()

        if not self.file_exists():
            raise exceptions.FileNotFoundError(self)

        old_name = self.name

        signals.pre_rename_file.send(
            sender=type(self),
            instance=self,
            old_name=old_name,
            new_name=new_name,
            options=options,
        )

        response = self._rename_file(new_name, **options)

        self.modified_at = now()

        signals.post_rename_file.send(
            sender=type(self),
            instance=self,
            old_name=old_name,
            new_name=new_name,
            options=options,
            response=response,
        )

    def _rename_file(self, new_name: str, **options):
        raise NotImplementedError

    def delete_file(self, **options):
        """
        Удаление файла.
        В действительности, удаление файла происходит в методе `_delete_file`.
        Не переопределяйте этот метод, если не уверены в том, что вы делаете.
        """
        signals.pre_delete_file.send(sender=type(self), instance=self, options=options)
        response = self._delete_file(**options)
        signals.post_delete_file.send(sender=type(self), instance=self, options=options, response=response)

    def _delete_file(self, **options):
        raise NotImplementedError


class FileFieldResource(FileFieldProxyMixin, FileResource):
    """
    Подкласс файлового ресурса, доступ к которому осуществляется через Django Storage.
    """

    class Meta(FileResource.Meta):
        abstract = True

    @property
    def name(self) -> str:
        self._require_file()
        return self.get_file().name

    def get_file_folder(self) -> str:
        """
        Возвращает путь к папке, в которую будет сохранен файл.
        Результат вызова используется в параметре `upload_to` в случае Django storage.
        """
        return ""

    def get_file(self) -> FieldFile:
        raise NotImplementedError

    def get_file_size(self) -> int:
        self._require_file()
        return self.get_file().size

    def get_file_url(self) -> str:
        self._require_file()
        return self.get_file().url

    def file_exists(self) -> bool:
        file = self.get_file()
        if not file:
            return False
        return file.storage.exists(file.name)

    def _attach_file(self, file: File, **options):
        self.get_file().save(file.name, file, save=False)

        self.basename = helpers.get_filename(file.name)
        self.extension = helpers.get_extension(self.name)

    def _rename_file(self, new_name: str, **options):
        file = self.get_file()
        with file.open() as fp:
            file.save(new_name, fp, save=False)

        self.basename = helpers.get_filename(new_name)
        self.extension = helpers.get_extension(self.name)

    def _delete_file(self, **options):
        self.get_file().delete(save=False)


class ImageFileResourceMixin(models.Model):
    """
    Подкласс файлового ресурса для изображений
    """

    title = models.CharField(
        _("title"),
        max_length=255,
        blank=True,
        help_text=_(
            "The title is being used as a tooltip when the user hovers the mouse over the image"
        ),
    )
    description = models.TextField(
        _("description"),
        blank=True,
        help_text=_(
            "This text will be used by screen readers, search engines, or when the image cannot be loaded"
        ),
    )
    width = models.PositiveSmallIntegerField(_("width"), default=0, editable=False)
    height = models.PositiveSmallIntegerField(_("height"), default=0, editable=False)
    cropregion = models.CharField(
        _("crop region"), max_length=24, blank=True, editable=False
    )

    class Meta:
        abstract = True

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),  # noqa
            "width": self.width,
            "height": self.height,
            "cropregion": self.cropregion,
            "title": self.title,
            "description": self.description,
        }

    def _prepare_file(self, file: File, **options) -> File:
        try:
            image = Image.open(file)
        except OSError:
            raise exceptions.UnsupportedFileError(
                _("File `%s` is not an image") % file.name
            )
        else:
            self.width, self.height = image.size

            # format extension
            file.name = replace_extension(file.name, format=image.format)
            root, ext = os.path.splitext(file.name)
            ext = ext.lstrip(".").lower()
            file.name = ".".join([root, ext])

        return super()._prepare_file(file, **options)  # noqa: F821


class VersatileImageResourceMixin(ImageFileResourceMixin):
    """
    Подкласс файлового ресурса для вариативного изображения.
    """

    # класс файла вариации
    variation_class = VariationFile

    # флаг, запускающий нарезку вариаций после сохранения экземпляра модели в БД
    need_recut = False

    class Meta(ImageFileResourceMixin.Meta):
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # нельзя использовать `_reset_variation_files`, т.к. он обращается
        # к `get_variations`, что может быть неопределено для элементов коллекций.
        self._variation_files_cache = {}

        # инициализация аттрибутов вариаций для уже загруженных файлов
        if self.pk:
            self._setup_variation_files()

    def __getattr__(self, item):
        # реализация-заглушка, чтобы PyCharm не ругался на атрибуты-вариации
        raise AttributeError(
            "{!r} object has no attribute {!r}".format(self.__class__.__name__, item)
        )

    def _reset_variation_files(self):
        self._variation_files_cache = {}
        for vname in self.get_variations():
            if vname in self.__dict__:
                del self.__dict__[vname]

    def _setup_variation_files(self):
        self._variation_files_cache.clear()
        for vname, vfile in self.variation_files():
            self.__dict__[vname] = vfile

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.need_recut:
            self.need_recut = False
            if settings.RQ_ENABLED:
                self.recut_async()
            else:
                self.recut()

    def attach_file(self, file: FileLike, name: str = None, **options):
        super().attach_file(file, name=name, **options)  # noqa: F821
        self.need_recut = True
        self._setup_variation_files()

    def _delete_file(self):
        self.delete_variations()
        super()._delete_file()  # noqa: F821

    def _rename_file(self, new_name: str, **options):
        super()._rename_file(new_name)  # noqa: F821
        self.recut()
        self._setup_variation_files()

    def get_variations(self) -> Dict[str, PaperVariation]:
        raise NotImplementedError

    def delete_variations(self):
        for vname, vfile in self.variation_files():
            vfile.delete()
        self._reset_variation_files()

    def variation_files(self) -> Iterable[Tuple[str, VariationFile]]:
        if not self._variation_files_cache:
            if not self.get_file():
                return

            self._variation_files_cache = {
                vname: self.get_variation_file(vname) for vname in self.get_variations()
            }
        yield from self._variation_files_cache.items()

    def get_variation_file(self, variation_name: str) -> VariationFile:
        return self.variation_class(instance=self, variation_name=variation_name)

    def calculate_max_size(self, source_size: Size) -> Optional[Tuple[int, int]]:
        """
        Вычисление максимально возможных значений ширины и высоты изображения
        среди всех вариаций, чтобы передать их в Image.draft().
        """
        max_width = 0
        max_height = 0
        for name, variation in self.get_variations().items():
            size = variation.get_output_size(source_size)
            max_width = max(max_width, size[0])
            max_height = max(max_height, size[1])
        if max_width and max_height:
            return max_width, max_height

    def _save_variation(self, name: str, variation: PaperVariation, image: Image):
        """
        Запись изображения в файловое хранилище
        """
        variation_file = self.get_variation_file(name)
        with variation_file.open("wb") as fp:
            variation.save(image, fp)

    def recut(self, names: Iterable[str] = ()):
        """
        Нарезка вариаций.
        Можно указать имена конкретных вариаций в параметре `names`.
        """
        if not self.file_exists():
            raise exceptions.FileNotFoundError(self)

        file = self.get_file()
        with file.open() as source:
            img = Image.open(source)
            draft_size = self.calculate_max_size(img.size)
            img = prepare_image(img, draft_size=draft_size)

            for name, variation in self.get_variations().items():
                if names and name not in names:
                    continue

                image = variation.process(img)
                self._save_variation(name, variation, image)

                signals.variation_created.send(
                    sender=type(self), instance=self, file=self.get_variation_file(name)
                )

    def recut_async(self, names: Iterable[str] = ()):
        """
        Добавление задачи нарезки вариаций в django-rq.
        """
        from django_rq.queues import get_queue

        queue = get_queue(settings.RQ_QUEUE_NAME)
        queue.enqueue_call(
            self._recut_task,
            kwargs={
                "app_label": self._meta.app_label,
                "model_name": self._meta.model_name,
                "object_id": self.pk,
                "using": self._state.db,
                "names": names,
            },
        )

    @classmethod
    def _recut_task(
        cls,
        app_label: str,
        model_name: str,
        object_id: int,
        using: str,
        names: Iterable[str],
    ):
        """
        Задача для django-rq.
        Вызывает `recut()` экземпляра в отдельном процессе.
        """
        instance = helpers.get_instance(app_label, model_name, object_id, using=using)
        instance.recut(names)



from django.urls import path

from .. import views
from .base import UploadedFileBase


class CollectionAdminBase(UploadedFileBase):
    create_collection_view_class = views.collection.CreateCollectionView
    delete_collection_view_class = views.collection.DeleteCollectionView
    upload_item_view_class = views.collection.UploadFileView
    delete_item_view_class = views.collection.DeleteFileView
    change_item_view_class = views.collection.ChangeFileView
    sort_items_view_class = views.collection.SortItemsView

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        urlpatterns = [
            path(
                "create/",
                self.admin_site.admin_view(self.create_collection_view_class.as_view()),
                name="%s_%s_create" % info,
            ),
            path(
                "delete/",
                self.admin_site.admin_view(self.delete_collection_view_class.as_view()),
                name="%s_%s_delete" % info,
            ),
            path(
                "upload_item/",
                self.admin_site.admin_view(self.upload_item_view_class.as_view()),
                name="%s_%s_upload_item" % info,
            ),
            path(
                "delete_item/",
                self.admin_site.admin_view(self.delete_item_view_class.as_view()),
                name="%s_%s_delete_item" % info,
            ),
            path(
                "change_item/",
                self.admin_site.admin_view(self.change_item_view_class.as_view()),
                name="%s_%s_change_item" % info,
            ),
            path(
                "sort_items/",
                self.admin_site.admin_view(self.sort_items_view_class.as_view()),
                name="%s_%s_sort_items" % info,
            ),
        ]
        return urlpatterns



from django import forms
from django.urls import reverse_lazy

from ...conf import settings
from .base import FileWidgetBase
from .mixins import FileUploaderWidgetMixin


class CollectionWidget(FileUploaderWidgetMixin, FileWidgetBase):
    template_name = "paper_uploads/collection_widget.html"

    @property
    def media(self):
        return forms.Media(
            js=[
                "paper_uploads/dist/widget.js",
            ],
            css={
                "screen": [
                    "paper_uploads/dist/widget.css",
                ],
            },
        )

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context.update(
            {
                "collection_cls": self.model,
                "preview_width": settings.COLLECTION_ITEM_PREVIEW_WIDTH,
                "preview_height": settings.COLLECTION_ITEM_PREVIEW_HEIGTH,
            }
        )

        # urls
        info = self.model._meta.app_label, self.model._meta.model_name
        context.update(
            {
                "create_collection_url": reverse_lazy("admin:%s_%s_create" % info),
                "delete_collection_url": reverse_lazy("admin:%s_%s_delete" % info),
                "upload_item_url": reverse_lazy("admin:%s_%s_upload_item" % info),
                "change_item_url": reverse_lazy("admin:%s_%s_change_item" % info),
                "delete_item_url": reverse_lazy("admin:%s_%s_delete_item" % info),
                "sort_items_url": reverse_lazy("admin:%s_%s_sort_items" % info),
            }
        )
        return context

    def get_instance(self, value):
        return self.model._base_manager.prefetch_related("items").get(pk=value)



from django.contrib import admin

from ..admin.file import UploadedFileAdmin
from ..admin.image import UploadedImageAdmin
from .models import CloudinaryFile, CloudinaryImage, CloudinaryMedia

admin.register(CloudinaryFile)(UploadedFileAdmin)
admin.register(CloudinaryImage)(UploadedImageAdmin)
admin.register(CloudinaryMedia)(UploadedFileAdmin)



from django.apps import apps
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from ...models.base import FileResource, VersatileImageResourceMixin
from ...models.collection import Collection, CollectionItemBase
from ...models.mixins import BacklinkModelMixin


class Command(BaseCommand):
    help = """
    Проверка целостности экземпляров файловых моделей.
    
    Если указан параметр `--fix-missing`, недостающие файлы вариаций
    создаются из исходного изображения.
    """
    options = None
    verbosity = None
    database = DEFAULT_DB_ALIAS

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )
        parser.add_argument(
            "--fix-missing",
            action="store_true",
            default=False,
            help="Recreate all missing variation files from a source.",
        )

    def check_exists(self):
        for model in apps.get_models():
            if not issubclass(model, FileResource):
                continue

            total = model._base_manager.using(self.database).count()
            if not total:
                continue

            queryset = model._base_manager.using(self.database)
            for index, instance in enumerate(queryset.iterator(), start=1):
                if self.verbosity >= 2:
                    self.stdout.write("\r" + (" " * 80), ending="\r")

                invalid = False
                message = "Errors were found in '{}.{}' #{instance.pk} ({instance}):".format(
                    model._meta.app_label, model._meta.model_name, instance=instance
                )

                if not instance.file_exists():
                    invalid = True
                    message += "\n  Not found source file"

                if invalid:
                    self.stdout.write(self.style.ERROR(message))

                if self.verbosity >= 2:
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Check file existence of '{}.{}' ({}/{}) ...\r".format(
                                model._meta.app_label,
                                model._meta.model_name,
                                index,
                                total,
                            )
                        ),
                        ending="",
                    )

            if self.verbosity >= 2:
                self.stdout.write("")

    def check_variations(self):
        for model in apps.get_models():
            if not issubclass(model, VersatileImageResourceMixin):
                continue

            total = model._base_manager.using(self.database).count()
            if not total:
                continue

            for index, instance in enumerate(
                model._base_manager.using(self.database).iterator(), start=1
            ):
                assert isinstance(instance, VersatileImageResourceMixin)

                if self.verbosity >= 2:
                    self.stdout.write("\r" + (" " * 80), ending="\r")

                invalid = False
                message = "Errors were found in '{}.{}' #{instance.pk} ({instance}):".format(
                    model._meta.app_label, model._meta.model_name, instance=instance
                )

                missed_variations = []
                for vname, vfile in instance.variation_files():
                    if vfile is not None and not vfile.exists():
                        missed_variations.append(vname)

                if missed_variations:
                    invalid = True
                    recreatable = self.options["fix_missing"] and instance.file_exists()
                    for vname in missed_variations:
                        message += "\n  Not found variation '{}'".format(vname)
                        if recreatable:
                            message += self.style.SUCCESS(" (recreated)")

                    if recreatable:
                        instance.recut(names=missed_variations)

                if invalid:
                    self.stdout.write(self.style.ERROR(message))

                if self.verbosity >= 2:
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Check variation existence of '{}.{}' ({}/{}) ...\r".format(
                                model._meta.app_label,
                                model._meta.model_name,
                                index,
                                total,
                            )
                        ),
                        ending="",
                    )

            if self.verbosity >= 2:
                self.stdout.write("")

    def check_owners(self):
        for model in apps.get_models():
            if not issubclass(model, BacklinkModelMixin):
                continue

            if model._meta.proxy:
                continue

            total = model._base_manager.using(self.database).count()
            if not total:
                continue

            for index, instance in enumerate(
                model._base_manager.using(self.database).iterator(), start=1
            ):
                assert isinstance(instance, BacklinkModelMixin)

                if self.verbosity >= 2:
                    self.stdout.write("\r" + (" " * 80), ending="\r")

                real_model = model
                if isinstance(instance, Collection):
                    real_model = instance.collection_content_type.model_class()

                invalid = False
                message = "Errors were found in '{}.{}' #{instance.pk} ({instance}):".format(
                    real_model._meta.app_label,
                    real_model._meta.model_name,
                    instance=instance,
                )

                owner_model = instance.get_owner_model()
                if owner_model is None:
                    invalid = True
                    message += "\n  Owner model '{}.{}' doesn't exists".format(
                        instance.owner_app_label, instance.owner_model_name,
                    )
                else:
                    owner_field = instance.get_owner_field()
                    if owner_field is None:
                        invalid = True
                        message += "\n  Owner model '{}.{}' has no field named '{}'".format(
                            instance.owner_app_label,
                            instance.owner_model_name,
                            instance.owner_fieldname,
                        )
                    else:
                        try:
                            owner_model._base_manager.get(
                                (instance.owner_fieldname, instance.pk)
                            )
                        except owner_model.DoesNotExist:
                            invalid = True
                            message += "\n  Owner instance '{}.{}' not found".format(
                                instance.owner_app_label, instance.owner_model_name,
                            )

                if invalid:
                    self.stdout.write(self.style.ERROR(message))

                if self.verbosity >= 2:
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Check owner of '{}.{}' ({}/{}) ...\r".format(
                                model._meta.app_label,
                                model._meta.model_name,
                                index,
                                total,
                            )
                        ),
                        ending="",
                    )

            if self.verbosity >= 2:
                self.stdout.write("")

    def check_item_types(self):
        total = CollectionItemBase.objects.using(self.database).count()
        for index, item in enumerate(
            CollectionItemBase.objects.using(self.database).iterator(), start=1
        ):
            assert isinstance(item, CollectionItemBase)

            if self.verbosity >= 2:
                self.stdout.write("\r" + (" " * 80), ending="\r")

            invalid = False
            message = "Errors were found in '{}.{}' #{item.pk} ({item}):".format(
                item._meta.app_label, item._meta.model_name, item=item
            )

            collection_cls = item.get_collection_class()
            if item.item_type not in collection_cls.item_types:
                invalid = True
                message += "\n  Item type '{}' is not defined in collection '{}.{}' #{}".format(
                    item.item_type,
                    collection_cls._meta.app_label,
                    collection_cls._meta.model_name,
                    item.collection_id,
                )
            else:
                item_model = collection_cls.item_types[item.item_type].model
                if item_model is not type(item):
                    invalid = True
                    message += "\n  Item class '{}.{}' is different from the '{}.{}'".format(
                        item._meta.app_label,
                        item._meta.model_name,
                        collection_cls._meta.app_label,
                        collection_cls._meta.model_name,
                    )

            if invalid:
                self.stdout.write(self.style.ERROR(message))

            if self.verbosity >= 2:
                self.stdout.write(
                    self.style.SUCCESS(
                        "Check item_type of collection items ({}/{}) ...\r".format(
                            index, total
                        )
                    ),
                    ending="",
                )

        if self.verbosity >= 2:
            self.stdout.write("")

    def handle(self, *args, **options):
        self.options = options
        self.verbosity = options["verbosity"]
        self.database = options["database"]

        self.check_item_types()
        self.check_owners()
        self.check_exists()
        self.check_variations()



from django.db.models.signals import Signal

# изменение контрольной суммы файла
checksum_update = Signal(providing_args=["instance", "checksum"])

# создание / перезапись (recut) файла вариации изображения
variation_created = Signal(providing_args=["instance", "file"])

# присоединение файла к ресурсу
pre_attach_file = Signal(providing_args=["instance", "file", "options"])
post_attach_file = Signal(providing_args=["instance", "file", "options", "response"])

# переименование файла
pre_rename_file = Signal(providing_args=["instance", "old_name", "new_name", "options"])
post_rename_file = Signal(providing_args=["instance", "old_name", "new_name", "options", "response"])

# удаление файла
pre_delete_file = Signal(providing_args=["instance", "options"])
post_delete_file = Signal(providing_args=["instance", "options", "response"])

# изменен порядок элементов коллекции
collection_reordered = Signal(providing_args=["instance"])



class CloudinaryOptionsMixin:
    def __init__(self, *args, cloudinary=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.cloudinary = cloudinary

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if "cloudinary" in kwargs:
            del kwargs["cloudinary"]
        return name, path, args, kwargs



import posixpath
from collections import OrderedDict
from typing import Any, Dict, Iterable, Optional, Type

import magic
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.staticfiles.finders import find
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core import checks
from django.core.files import File
from django.db import models, transaction
from django.db.models import functions
from django.db.models.base import ModelBase
from django.db.models.fields.files import FieldFile
from django.template import loader
from django.utils.module_loading import import_string
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from polymorphic.base import PolymorphicModelBase
from polymorphic.models import PolymorphicModel

from ..conf import FILE_ICON_DEFAULT, FILE_ICON_OVERRIDES, settings
from ..helpers import _get_item_types, _set_item_types, build_variations
from ..storage import upload_storage
from ..variations import PaperVariation
from .base import (
    FileFieldResource,
    NoPermissionsMetaBase,
    ResourceBaseMeta,
    VersatileImageResourceMixin,
)
from .fields import CollectionItem
from .fields.collection import ContentItemRelation
from .image import VariationalFileField
from .mixins import BacklinkModelMixin
from .utils import generate_filename

__all__ = [
    "CollectionItemBase",
    "CollectionBase",
    "FilePreviewMixin",
    "FileItem",
    "SVGItem",
    "ImageItem",
    "Collection",
    "ImageCollection",
]


class ItemTypesDescriptor:
    """
    Дескриптор, добавляемый к моделям коллекций.
    Возвращает упорядоченый словарь типов элементов коллекции,
    учитывая наследование.
    """

    def __init__(self, name):
        self.name = name

    def __get__(self, instance, cls=None):
        if cls is None:
            cls = type(instance)

        cache_attname = "_{}__item_type_cache".format(cls.__name__)
        if hasattr(cls, cache_attname):
            return OrderedDict(getattr(cls, cache_attname))

        item_types = OrderedDict()
        parents = [base for base in cls.mro() if issubclass(base, CollectionBase)]
        for base in reversed(parents):
            base_item_types = _get_item_types(base)
            if base_item_types is not None:
                undefined_item_types = set(item_types.keys()).difference(
                    base_item_types
                )
                item_types.update(base_item_types)

                # delete overridden item types
                for item_type_name in undefined_item_types:
                    if hasattr(base, item_type_name):
                        del item_types[item_type_name]

        # cache result in class object
        setattr(cls, cache_attname, tuple(item_types.items()))
        return item_types


class CollectionMeta(NoPermissionsMetaBase, ModelBase):
    """
    Хак, при котором вместо наследования создаются прокси-модели,
    если явно не указано обратное.
    """

    @classmethod
    def __prepare__(cls, name, bases):
        return OrderedDict()

    def __new__(mcs, name, bases, attrs, **kwargs):  # noqa: N804
        # set proxy=True by default
        meta = attrs.pop("Meta", None)
        if meta is None:
            meta = type("Meta", (), {"proxy": True})
        else:
            meta_attrs = meta.__dict__.copy()
            meta_attrs.setdefault("proxy", True)
            meta = type("Meta", meta.__bases__, meta_attrs)
        attrs["Meta"] = meta

        # сохраняем явно объявленные (не унаследованные) типы элементов коллекции
        # в приватное поле класса
        item_types = OrderedDict()
        for key, value in list(attrs.items()):
            if isinstance(value, CollectionItem):
                item_types[key] = value

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)
        _set_item_types(new_class, item_types)
        new_class.item_types = ItemTypesDescriptor("item_types")
        return new_class


class CollectionBase(BacklinkModelMixin, metaclass=CollectionMeta):
    items = ContentItemRelation(
        "paper_uploads.CollectionItemBase",
        content_type_field="collection_content_type",
        object_id_field="collection_id",
        for_concrete_model=False,
    )
    created_at = models.DateTimeField(_("created at"), default=now, editable=False)

    class Meta:
        proxy = False
        abstract = True
        default_permissions = ()
        verbose_name = _("collection")
        verbose_name_plural = _("collections")

    def delete(self, using=None, keep_parents=False):
        # удаляем элементы вручную из-за рекурсии (см. clean_uploads.py, строка 145)
        self.items.all().delete()
        super().delete(using=using, keep_parents=keep_parents)

    @classmethod
    def _check_fields(cls, **kwargs):
        errors = super()._check_fields(**kwargs)
        for field in cls.item_types.values():
            errors.extend(field.check(**kwargs))
        return errors

    def get_items(self, item_type: str = None) -> 'models.QuerySet[CollectionItemBase]':
        # TODO: что если класс элемента был удален из коллекции, но элементы остались?
        if item_type is None:
            return self.items.order_by("order")
        if item_type not in self.item_types:
            raise ValueError(_("Unsupported collection item type: %s") % item_type)
        return self.items.filter(item_type=item_type).order_by("order")

    def detect_item_type(self, *args, **kwargs) -> Optional[str]:
        """
        Генератор, поочередно проверяющий классы элементов коллекции
        на возможность представления данных, переданных в параметрах.
        """
        for item_type, field in self.item_types.items():
            if hasattr(field.model, "file_supported"):
                if field.model.file_supported(*args, **kwargs):
                    yield item_type


class CollectionManager(models.Manager):
    """
    Из-за того, что все галереи являются прокси-моделями, запросы от имени
    любого из классов галереи затрагивает абсолютно все объекты галереи.
    С помощью этого менеджера можно работать только с галереями текущего типа.
    """

    def get_queryset(self):
        collection_ct = ContentType.objects.get_for_model(
            self.model, for_concrete_model=False
        )
        return super().get_queryset().filter(collection_content_type=collection_ct)


class Collection(CollectionBase):
    collection_content_type = models.ForeignKey(
        ContentType, null=True, on_delete=models.SET_NULL, editable=False
    )

    default_mgr = models.Manager()  # fix migrations manager
    objects = CollectionManager()

    class Meta:
        proxy = False  # явно указываем, что это не прокси-модель
        default_manager_name = "default_mgr"

    def save(self, *args, **kwargs):
        if not self.collection_content_type:
            self.collection_content_type = ContentType.objects.get_for_model(
                self, for_concrete_model=False
            )
        super().save(*args, **kwargs)


# ======================================================================================


class CollectionItemMetaBase(PolymorphicModelBase, ResourceBaseMeta):
    pass


class CollectionItemBase(PolymorphicModel, metaclass=CollectionItemMetaBase):
    """
    Базовый класс элемента коллекции.
    """

    # Флаг для индикации базового класса элемента коллекции.
    # См. метод _check_form_class()
    __BaseCollectionItem = True

    change_form_class: Optional[str] = None

    # путь к шаблону, представляющему элемент коллекции в админке
    template_name: Optional[str] = None

    # путь к шаблону, представляющему картинку-превью элемента коллекции в админке
    preview_template_name: Optional[str] = None

    collection_content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    collection_id = models.IntegerField()
    collection = GenericForeignKey(
        ct_field="collection_content_type",
        fk_field="collection_id",
        for_concrete_model=False,
    )

    item_type = models.CharField(
        _("type"), max_length=32, db_index=True, editable=False
    )
    order = models.IntegerField(_("order"), default=0, editable=False)

    class Meta:
        verbose_name = _("item")
        verbose_name_plural = _("items")

    @classmethod
    def check(cls, **kwargs):
        return [
            *super().check(**kwargs),
            *cls._check_form_class(),
            *cls._check_template_name(),
        ]

    @classmethod
    def _check_form_class(cls, **kwargs):
        flag = "_{}__BaseCollectionItem".format(cls.__name__)
        if getattr(cls, flag, None) is True or cls._meta.abstract:
            return []

        errors = []
        if cls.change_form_class is None:
            errors.append(
                checks.Error(
                    "{} requires a definition of 'change_form_class'".format(
                        cls.__name__
                    ),
                    obj=cls,
                )
            )
        else:
            try:
                import_string(cls.change_form_class)
            except ImportError:
                errors.append(
                    checks.Error(
                        "The value of 'change_form_class' refers to '%s', which does "
                        "not exists" % cls.change_form_class,
                        obj=cls,
                    )
                )
        return errors

    @classmethod
    def _check_template_name(cls, **kwargs):
        flag = "_{}__BaseCollectionItem".format(cls.__name__)
        if getattr(cls, flag, None) is True or cls._meta.abstract:
            return []

        errors = []
        if cls.template_name is None:
            errors.append(
                checks.Error(
                    "{} requires a definition of 'template_name'".format(cls.__name__),
                    obj=cls,
                )
            )
        return errors

    def save(self, *args, **kwargs):
        if not self.pk:
            # попытка решить проблему того, что при создании коллекции,
            # элементы отсортированы в порядке загрузки, а не в порядке
            # добавления. Код ниже не решает проблему, но уменьшает её влияние.
            if self.collection.items.filter(order=self.order).exists():
                max_order = self.collection.items.aggregate(
                    order=functions.Coalesce(models.Max("order"), 0)
                )["order"]
                self.order = max_order + 1
        super().save(*args, **kwargs)

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "collectionId": self.collection_id,
            "itemType": self.item_type,
            "caption": self.get_caption(),
            "order": self.order,
            "preview": self.render_preview(),
        }

    def get_collection_class(self) -> Type[CollectionBase]:
        return self.collection_content_type.model_class()

    def get_itemtype_field(self) -> Optional[CollectionItem]:
        collection_cls = self.get_collection_class()
        for name, field in collection_cls.item_types.items():
            if field.model is type(self):
                return field
        return None

    def attach_to(self, collection: CollectionBase):
        """
        Подключение элемента к коллекции.
        """
        self.collection_content_type = ContentType.objects.get_for_model(
            collection, for_concrete_model=False
        )
        self.collection_id = collection.pk
        for name, field in collection.item_types.items():
            if field.model is type(self):
                self.item_type = name
                break
        else:
            raise TypeError(_("Unsupported collection item: %s") % type(self).__name__)

    def get_caption(self):
        """ Заголовок для виджета в админке """
        return self.get_basename()

    def render_preview(self):
        """ Отображение элемента коллекции в админке """
        context = self.get_preview_context()
        return loader.render_to_string(self.preview_template_name, context)

    def get_preview_context(self):
        return {
            "item": self,
            "width": settings.COLLECTION_ITEM_PREVIEW_WIDTH,
            "height": settings.COLLECTION_ITEM_PREVIEW_HEIGTH,
        }


class CollectionFileItemBase(CollectionItemBase, FileFieldResource, metaclass=CollectionItemMetaBase):
    """
    Базовый класс элемента галереи, содержащего файл.
    """

    class Meta:
        abstract = True

    @classmethod
    def file_supported(cls, file: File) -> bool:
        """
        Проверка возможности представления загруженного файла
        текущим классом элемента в коллекции.
        """
        raise NotImplementedError


class FilePreviewMixin(models.Model):
    """
    Миксина элемента коллекции, добавляющая иконку для файла в админке
    """

    class Meta:
        abstract = True

    def get_preview_url(self):
        extension = self.extension.lower()  # noqa
        extension = FILE_ICON_OVERRIDES.get(extension, extension)
        icon_path_template = "paper_uploads/dist/assets/{}.svg"
        icon_path = icon_path_template.format(extension)
        if find(icon_path) is None:
            icon_path = icon_path_template.format(FILE_ICON_DEFAULT)
        return staticfiles_storage.url(icon_path)

    def get_preview_context(self):
        context = super().get_preview_context()  # noqa
        context.update(preview_url=self.get_preview_url())
        return context


# ======================================================================================


class FileItem(FilePreviewMixin, CollectionFileItemBase):
    change_form_class = "paper_uploads.forms.dialogs.collection.FileItemDialog"
    template_name = "paper_uploads/items/file.html"
    preview_template_name = "paper_uploads/items/preview/file.html"

    file = models.FileField(
        _("file"),
        max_length=255,
        storage=upload_storage,
        upload_to=generate_filename,
    )
    display_name = models.CharField(_("display name"), max_length=255, blank=True)

    class Meta(CollectionItemBase.Meta):
        verbose_name = _("File item")
        verbose_name_plural = _("File items")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.basename
        super().save(*args, **kwargs)

    def get_file_folder(self) -> str:
        return settings.COLLECTION_FILES_UPLOAD_TO

    def get_file(self) -> FieldFile:
        return self.file

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> models.FileField:
        return self._meta.get_field("file")

    @classmethod
    def file_supported(cls, file: File) -> bool:
        return True


class SVGItem(CollectionFileItemBase):
    change_form_class = "paper_uploads.forms.dialogs.collection.FileItemDialog"
    template_name = "paper_uploads/items/svg.html"
    preview_template_name = "paper_uploads/items/preview/svg.html"

    file = models.FileField(
        _("file"),
        max_length=255,
        storage=upload_storage,
        upload_to=generate_filename,
    )
    display_name = models.CharField(_("display name"), max_length=255, blank=True)

    class Meta(CollectionItemBase.Meta):
        verbose_name = _("SVG item")
        verbose_name_plural = _("SVG items")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.basename
        super().save(*args, **kwargs)

    def get_file_folder(self) -> str:
        return settings.COLLECTION_FILES_UPLOAD_TO

    def get_file(self) -> FieldFile:
        return self.file

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> models.FileField:
        return self._meta.get_field("file")

    @classmethod
    def file_supported(cls, file: File) -> bool:
        filename, ext = posixpath.splitext(file.name)
        return ext.lower() == ".svg"


class ImageItem(VersatileImageResourceMixin, CollectionFileItemBase):
    PREVIEW_VARIATIONS = settings.COLLECTION_IMAGE_ITEM_PREVIEW_VARIATIONS
    change_form_class = "paper_uploads.forms.dialogs.collection.ImageItemDialog"
    template_name = "paper_uploads/items/image.html"
    preview_template_name = "paper_uploads/items/preview/image.html"

    file = VariationalFileField(
        _("file"),
        max_length=255,
        storage=upload_storage,
        upload_to=generate_filename,
    )

    class Meta(CollectionItemBase.Meta):
        verbose_name = _("Image item")
        verbose_name_plural = _("Image items")

    def get_file_folder(self) -> str:
        return settings.COLLECTION_IMAGES_UPLOAD_TO

    def get_file(self) -> FieldFile:
        return self.file

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> VariationalFileField:
        return self._meta.get_field("file")

    def recut_async(self, names: Iterable[str] = ()):
        """
        Превью для админки режутся сразу, а остальное — потом.
        """
        preview_variations = tuple(self.PREVIEW_VARIATIONS.keys())
        self._recut_sync(names=preview_variations)

        names = tuple(set(names).difference(preview_variations))
        super().recut_async(names)

    def get_variations(self) -> Dict[str, PaperVariation]:
        """
        Перебираем возможные места вероятного определения вариаций и берем
        первое непустое значение. Порядок проверки:
            1) параметр `variations` псевдо-поля `CollectionItem`
            2) член класса галереи VARIATIONS
        К найденному словарю примешиваются вариации для админки.
        """
        if not hasattr(self, "_variations_cache"):
            collection_cls = self.get_collection_class()
            itemtype_field = self.get_itemtype_field()
            variation_config = self.get_variation_config(itemtype_field, collection_cls)
            self._variations_cache = build_variations(variation_config)
        return self._variations_cache

    @classmethod
    def file_supported(cls, file: File) -> bool:
        mimetype = magic.from_buffer(file.read(1024), mime=True)
        file.seek(0)  # correct file position after mimetype detection
        basetype, subtype = mimetype.split("/", 1)
        return basetype == "image"

    @classmethod
    def get_variation_config(
        cls, rel: Optional[CollectionItem], collection_cls: Type[CollectionBase]
    ) -> Dict[str, Any]:
        if rel is not None and "variations" in rel.options:
            variations = rel.options["variations"]
        else:
            variations = getattr(collection_cls, "VARIATIONS", None)
        variations = (variations or {}).copy()
        variations.update(cls.PREVIEW_VARIATIONS)
        return variations


# ==============================================================================


class ImageCollection(Collection):
    """
    Коллекция, позволяющая хранить только изображения.
    """

    image = CollectionItem(ImageItem)

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "strictImageValidation": True,
            "acceptFiles": [
                "image/bmp",
                "image/gif",
                "image/jpeg",
                "image/png",
                # "image/svg+xml",
                "image/tiff",
                "image/webp",
            ],
        }



class InvalidRequest(Exception):
    pass


class InvalidContentType(InvalidRequest):
    pass


class InvalidObjectId(InvalidRequest):
    pass


class InstanceNotFound(InvalidRequest):
    pass


class InvalidChunking(InvalidRequest):
    pass


class InvalidUUID(InvalidRequest):
    def __init__(self, value):
        self.value = value
        super().__init__()


class ContinueUpload(Exception):
    pass


class UncompleteUpload(Exception):
    pass


class AjaxFormError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(message)


class UnsupportedFileError(Exception):
    """
    Исключение, вызываемое внутри метода `attach_file` при обнаружении
    ситуации, когда файл не может быть представлен текущей моделью.
    """
    def __init__(self, message):
        self.message = message
        super().__init__(message)


class FileNotFoundError(Exception):
    def __init__(self, file):
        self.file = file
        self.name = file.name
        super().__init__()



import sys
from typing import List, Type, Union

from django.apps import apps
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS, models
from django.db.models.fields import Field

from ... import exceptions
from ...models.base import VersatileImageResourceMixin
from ...models.collection import Collection
from ...models.fields.collection import CollectionItem


def is_image(field: Field) -> bool:
    """
    Возвращает True, если поле ссылается на класс изображения с вариациями.
    """
    return field.is_relation and issubclass(
        field.related_model, VersatileImageResourceMixin
    )


def is_image_item(field: CollectionItem) -> bool:
    """
    Возвращает True, если поле коллекции подключает класс элемента
    изображения с вариациями.
    """
    return issubclass(field.model, VersatileImageResourceMixin)


def is_collection(model: Type[Union[models.Model, Collection]]) -> bool:
    """
    Возвращает True, если model - коллекция.
    """
    return issubclass(model, Collection)


def get_collection_variations(model: Type[Collection], item_type_class: CollectionItem) -> List[str]:
    return list(
        item_type_class.model.get_variation_config(item_type_class, model).keys()
    )


def get_field_variations(field: VersatileImageResourceMixin) -> List[str]:
    return list(
        field.variations.keys()
    )


class Command(BaseCommand):
    help = """
    Создание/перезапись вариаций для указанной модели из исходного изображения.
    """
    verbosity = None
    database = DEFAULT_DB_ALIAS

    def add_arguments(self, parser):
        parser.add_argument(
            "model",
            metavar="app_label.ModelName",
            help="Specifies the model to recreate variations for",
        )
        parser.add_argument(
            "--field",
            nargs="?",
            help="Restricts variations to the specified field. "
                 "You should not specify any field for Collection models.",
        )
        parser.add_argument(
            "--item-type",
            nargs="?",
            help="Only look for variations in the specified CollectionItem. "
                 "Use this argument for Collection models only.",
        )
        parser.add_argument(
            "--variations",
            dest="variations",
            nargs="+",
            help="Specifies the variation names to recreate",
        )
        parser.add_argument(
            "-i", "--interactive",
            action="store_true",
        )
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )

    def handle(self, *args, **options):
        self.verbosity = options["verbosity"]
        self.database = options["database"]

        model_name = options["model"]
        model = apps.get_model(model_name)
        if is_collection(model):
            item_type = options["item_type"]
            if not item_type:
                raise RuntimeError("The argument 'item-type' required")

            if item_type not in model.item_types:
                raise RuntimeError("Unsupported collection item type: %s" % item_type)

            item_type_class = model.item_types[item_type]
            if not is_image_item(item_type_class):
                raise RuntimeError("Not an image or invalid collection item type: %s" % item_type)

            variations = options["variations"]
            if not variations:
                if options["interactive"]:
                    variations = self.select_variations_dialog(model, item_type_class)

            if not variations:
                variations = set(get_collection_variations(model, item_type_class))

            self.process_collection(model, item_type, variations)
        else:
            fieldname = options["field"]
            if not fieldname:
                raise RuntimeError("The argument 'field' required")

            field = model._meta.get_field(fieldname)
            if not is_image(field):
                raise RuntimeError("Not an image or invalid field: %s" % fieldname)

            variations = options["variations"]
            if not variations:
                if options["interactive"]:
                    variations = self.select_variations_dialog(model, field)

            if not variations:
                variations = set(get_field_variations(field))

            self.process_field(model, fieldname, variations)

    def process_collection(self, model, item_type, variations):
        total = model.objects.using(self.database).count()  # use `objects` manager!
        for index, collection in enumerate(
            model.objects.using(self.database).iterator(), start=1
        ):
            if self.verbosity >= 1:
                self.stdout.write(
                    self.style.SUCCESS(
                        "Recreate variations for `{}` #{} ({}/{}) ... ".format(
                            model.__name__, collection.pk, index, total
                        )
                    )
                )

            for item in collection.get_items(item_type).using(self.database).iterator():
                try:
                    item.recut(names=variations)
                except exceptions.FileNotFoundError:
                    self.stderr.write(
                        "Not found source file for `{}` #{}".format(
                            item.__class__.__name__, item.pk
                        )
                    )

    def process_field(self, model, fieldname, variations):
        instances = model._base_manager.using(self.database).exclude(
            (fieldname, None)
        )
        total = instances.count()
        for index, instance in enumerate(instances.iterator(), start=1):
            if self.verbosity >= 1:
                self.stdout.write(
                    self.style.SUCCESS(
                        "Recreate variations for `{}` #{} ({}/{}) ... ".format(
                            model.__name__, instance.pk, index, total
                        )
                    ),
                    ending="\r",
                )
            field = getattr(instance, fieldname)

            try:
                field.recut(names=variations)
            except exceptions.FileNotFoundError:
                self.stderr.write(
                    "Not found source file for `{}` #{}".format(
                        model.__name__, instance.pk
                    )
                )

        self.stdout.write("")

    def select_variations_dialog(self, model, field):
        if is_collection(model):
            allowed_variations = get_collection_variations(model, field)
        else:
            allowed_variations = get_field_variations(field)

        while True:
            self.stdout.write(
                self.style.SUCCESS(
                    "Please, specify the variations you would like to process."
                )
            )
            self.stdout.write(" {}) All variations".format(self.style.SUCCESS("*")))
            for index, vname in enumerate(allowed_variations, start=1):
                self.stdout.write(
                    "{}) Variation `{}`".format(
                        self.style.SUCCESS("{:>2}".format(index)), vname
                    )
                )
            self.stdout.write(" {}) Abort".format(self.style.SUCCESS("q")))
            self.stdout.write("Enter your choice: ", ending="")

            answer_string = input().strip()
            answers = tuple(map(str.strip, answer_string.split(",")))
            if "*" in answers:
                return allowed_variations
            if "q" in answers:
                sys.exit()

            try:
                answers = tuple(map(int, answers))
            except ValueError:
                self.stderr.write(
                    "Invalid selection. Press Enter to try again... ", ending=""
                )
                input()
                continue

            if not all(1 <= answer <= len(allowed_variations) for answer in answers):
                self.stderr.write(
                    "Invalid selection. Press Enter to try again... ", ending=""
                )
                input()
                continue

            return set(allowed_variations[answer - 1] for answer in answers)



from django import forms
from django.contrib.admin import site
from django.contrib.admin.sites import NotRegistered
from django.urls import reverse_lazy

from .base import FileWidgetBase
from .mixins import FileUploaderWidgetMixin


class ImageWidget(FileUploaderWidgetMixin, FileWidgetBase):
    template_name = "paper_uploads/image_widget.html"

    @property
    def media(self):
        return forms.Media(
            js=[
                "paper_uploads/dist/widget.js",
            ],
            css={
                "screen": [
                    "paper_uploads/dist/widget.css",
                ],
            },
        )

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        # support proxy models
        for model_class in {self.model, self.model._meta.concrete_model}:
            if site.is_registered(model_class):
                info = model_class._meta.app_label, model_class._meta.model_name
                context.update(
                    {
                        "upload_url": reverse_lazy("admin:%s_%s_upload" % info),
                        "change_url": reverse_lazy("admin:%s_%s_change" % info),
                        "delete_url": reverse_lazy("admin:%s_%s_delete" % info),
                    }
                )
                break
        else:
            raise NotRegistered("The model %s is not registered." % self.model.__name__)

        return context



from cloudinary.templatetags import cloudinary

from .models.base import CloudinaryFileResource


def paper_cloudinary_url(context, source, options_dict=None, **options):
    if isinstance(source, CloudinaryFileResource):
        source = source.get_file().resource
    return cloudinary.cloudinary_url(context, source, options_dict, **options)



from django import forms

from ...models.collection import FileItem, ImageItem
from .base import UploadedFileBaseForm


class FileItemDialog(UploadedFileBaseForm):
    class Meta:
        model = FileItem
        fields = ("new_name", "display_name")


class ImageItemDialog(UploadedFileBaseForm):
    class Meta:
        model = ImageItem
        fields = ("new_name", "title", "description")
        widgets = {
            "description": forms.Textarea(attrs={
                "rows": 3
            })
        }



from .file import CloudinaryFileField
from .image import CloudinaryImageField
from .media import CloudinaryMediaField

__all__ = [
    "CloudinaryFileField",
    "CloudinaryImageField",
    "CloudinaryMediaField",
]



from django import forms

from ...models.image import UploadedImage
from .base import UploadedFileBaseForm


class UploadedImageDialog(UploadedFileBaseForm):
    class Meta:
        model = UploadedImage
        fields = ("new_name", "title", "description")
        widgets = {
            "description": forms.Textarea(attrs={
                "rows": 3
            })
        }



from datetime import timedelta

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS, models, transaction
from django.utils.timezone import now
from polymorphic.models import PolymorphicModel

from ...models.base import FileResource
from ...models.collection import CollectionBase


class Command(BaseCommand):
    help = """
    Поиск и удаление экземпляров файловых моделей с утерянными файлами.
    Также удаляются экземпляров файловых моделей, на которые нет ссылок.
    
    Создание экземпляра файловой модели и загрузка в неё файла - это 
    две отдельные операции. Между ними может пройти какое-то время.
    Для того, чтобы сохранить экземпляры, в которые ещё не загрузились файлы,
    эта команда пропускает те экземпляры, которые созданы раньше, чем
    `--min-age` минут назад. Значение параметра `--min-age` по умолчанию 
    установлено в 30 минут.
    """
    verbosity = None
    database = DEFAULT_DB_ALIAS
    interactive = True

    def add_arguments(self, parser):
        parser.add_argument(
            "--min-age",
            type=int,
            default=30,
            help="Minimum instance age in minutes to look for",
        )
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            help="Do NOT prompt the user for input of any kind.",
        )

    @staticmethod
    def search_fields(related_model):
        """
        Поиск моделей и полей, ссылающихся на заданную.

        :type related_model: type
        :rtype: list of (models.Model, list)
        """
        for model in apps.get_models():
            if not model._meta.managed or model._meta.proxy:
                continue

            fields = []
            for field in model._meta.get_fields():
                if not field.concrete:
                    continue

                if not field.is_relation:
                    continue

                if issubclass(field.related_model, related_model):
                    fields.append(field)

            if fields:
                yield model, fields

    def get_used_ids(self, related_model, exclude_models=None):
        used_ids = set()
        for model, fields in self.search_fields(related_model):
            if exclude_models and issubclass(model, exclude_models):
                continue

            with transaction.atomic(self.database):
                for field in fields:
                    used_values = (
                        model._base_manager.using(self.database)
                        .exclude(models.Q((field.name, None)))
                        .values_list(field.name, flat=True)
                    )
                    used_ids.update(used_values)
        return used_ids

    def clean_model(self, queryset, exclude_models=None):
        """
        Поиск экземпляров, на которые нет ссылок с других моделей.
        """
        related_model = queryset.model
        used_ids = self.get_used_ids(related_model, exclude_models)

        unused_qs = queryset.exclude(pk__in=used_ids)
        unused_count = unused_qs.count()

        if not unused_count:
            if self.verbosity >= 2:
                self.stderr.write(
                    self.style.NOTICE(
                        "There are no unused instances of {}".format(
                            related_model.__name__
                        )
                    )
                )
            return

        if self.interactive:
            while True:
                answer = input(
                    "Found \033[92m%d unused %s\033[0m objects. \n"
                    "IDs: %s\n"
                    "What would you like to do with them?\n"
                    "(p)rint / (k)eep / (d)elete [default=keep]? "
                    % (
                        unused_count,
                        related_model.__name__,
                        ", ".join(map(str, unused_qs.values_list("pk", flat=True))),
                    )
                )
                answer = answer.lower() or "k"
                if answer in {"p", "print"}:
                    self.stdout.write("\n")
                    qs = unused_qs.order_by("pk")
                    for index, item in enumerate(qs, start=1):
                        self.stdout.write(
                            "  {}) {} #{} (File: {})".format(
                                index,
                                type(item).__name__,
                                item.pk,
                                item.name,
                            )
                        )
                    self.stdout.write("\n")
                elif answer in {"k", "keep"}:
                    return
                elif answer in {"d", "delete"}:
                    # `unused_qs.delete()` не используется по причине зацикленной рекурсии.
                    # Полиморфная модель ImageItem при инициализации добавляет в себя
                    # ссылки на файлы вариаций. Для этого требуется обращение к полю `file`.
                    # Но при удалении *коллекции* этого поля в модели нет, т.к. для SQL
                    # DELETE оно не нужно. Из-за этого вызывается метод
                    # `refresh_from_db`, который снова инициализирует модель ImageItem.
                    with transaction.atomic():
                        for obj in unused_qs:
                            obj.delete()
                    return

    def clean_source_missing(self, queryset):
        """
        Поиск экземпляров с утерянными файлами
        """
        sourceless_items = set()
        related_model = queryset.model
        for instance in queryset.iterator():
            if not instance.file_exists():
                sourceless_items.add(instance.pk)

        sourceless_qs = queryset.filter(pk__in=sourceless_items)
        sourceless_count = sourceless_qs.count()

        if not sourceless_count:
            if self.verbosity >= 2:
                self.stderr.write(
                    self.style.NOTICE(
                        "There are no instances of {} which have no source file.".format(
                            related_model.__name__
                        )
                    )
                )
            return

        if self.interactive:
            while True:
                answer = input(
                    "Found \033[92m%d %s\033[0m objects which are linked to a non-existent files.\n"
                    "IDs: %s\n"
                    "What would you like to do with them?\n"
                    "(p)rint / (k)eep / (d)elete [default=keep]? "
                    % (
                        sourceless_count,
                        related_model.__name__,
                        ", ".join(map(str, sourceless_qs.values_list("pk", flat=True))),
                    )
                )
                answer = answer.lower() or "k"
                if answer in {"p", "print"}:
                    self.stdout.write("\n")
                    qs = sourceless_qs.order_by("pk")
                    for index, item in enumerate(qs, start=1):
                        self.stdout.write(
                            "  {}) {} #{} (File: {})".format(
                                index,
                                type(item).__name__,
                                item.pk,
                                item.name,
                            )
                        )
                    self.stdout.write("\n")
                elif answer in {"k", "keep"}:
                    return
                elif answer in {"d", "delete"}:
                    sourceless_qs.delete()
                    return

    def handle(self, *args, **options):
        self.verbosity = options["verbosity"]
        self.database = options["database"]
        self.interactive = options["interactive"]

        min_age = now() - timedelta(minutes=options["min_age"])

        for model in apps.get_models():
            if not issubclass(model, FileResource):
                continue

            if issubclass(model, PolymorphicModel):
                continue

            if model._meta.abstract or model._meta.proxy:
                continue

            self.clean_source_missing(model._base_manager.using(self.database).all())
            self.clean_model(
                model._base_manager.using(self.database).filter(
                    uploaded_at__lte=min_age
                )
            )

        for model in apps.get_models():
            if not issubclass(model, FileResource):
                continue

            if not issubclass(model, PolymorphicModel):
                continue

            if model._meta.abstract:
                continue

            self.clean_source_missing(
                model.objects.using(self.database).non_polymorphic()
            )

        # Do not touch fresh galleries - they may not be saved yet.
        for model in apps.get_models():
            if issubclass(model, CollectionBase) and not model._meta.abstract:
                content_type = ContentType.objects.get_for_model(
                    model, for_concrete_model=False
                )
                collection_qs = model._base_manager.using(self.database).filter(
                    collection_content_type=content_type, created_at__lte=min_age
                )
                self.clean_model(collection_qs)



import datetime
import os
import posixpath
import tempfile
from io import UnsupportedOperation
from typing import Optional

import cloudinary.exceptions
import requests
from cloudinary import CloudinaryResource, uploader
from cloudinary.utils import upload_params
from django.core.exceptions import ValidationError
from django.core.files import File
from django.core.files.utils import FileProxyMixin
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from filelock import FileLock

from ... import exceptions, helpers, utils
from ...conf import settings
from ...logging import logger
from ...models.base import FileResource
from .mixins import ReadonlyCloudinaryFileProxyMixin


class CloudinaryFieldFile(FileProxyMixin):
    """
    Реализация интерфейса django.core.files.File для Cloudinary
    """
    DEFAULT_CHUNK_SIZE = 64 * 2 ** 10

    def __init__(self, resource: CloudinaryResource, checksum: str = None):
        self.resource = resource
        self.name = self.public_id
        self.checksum = checksum
        self.file = None

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self or "None")

    def __bool__(self):
        return bool(self.name)

    def __len__(self):
        return self.size

    def __eq__(self, other):
        return self.url == other.url

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    @cached_property
    def public_id(self):
        public_id = self.resource.public_id
        if (self.resource.resource_type == "raw") and self.resource.format:
            public_id += "." + self.resource.format
        return public_id

    @cached_property
    def metadata(self):
        if self.resource.metadata is not None:
            return self.resource.metadata

        return uploader.explicit(
            self.public_id,
            type=self.resource.type,
            resource_type=self.resource.resource_type
        )

    @property
    def url(self) -> str:
        return self.resource.url

    @property
    def size(self) -> int:
        return self.metadata["bytes"]

    @property
    def format(self) -> Optional[str]:
        return self.metadata.get("format")

    def _get_tempfile_path(self):
        return os.path.join(
            tempfile.gettempdir(),
            settings.CLOUDINARY_TEMP_DIR,
            self.name
        )

    def _download_file(self, mode="rb", chunk_size=None):
        """
        Скачивание файла из Cloudinary во временную директорию.
        Загруженный файл остается там даже после закрытия, чтобы не
        скачивать его заново в следующий раз.
        """
        chunk_size = chunk_size or self.DEFAULT_CHUNK_SIZE
        tempfile_path = self._get_tempfile_path()
        if os.path.exists(tempfile_path):
            with open(tempfile_path, "rb") as fp:
                file_checksum = utils.checksum(fp)
            if file_checksum == self.checksum:
                return File(open(tempfile_path, mode))

        root, basename = os.path.split(tempfile_path)
        os.makedirs(root, mode=0o755, exist_ok=True)

        lock = FileLock(tempfile_path + ".lock")
        with lock.acquire(timeout=3600):
            response = requests.get(self.url, stream=True)
            response.raise_for_status()
            with open(tempfile_path, "wb+") as fp:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    fp.write(chunk)
            return File(open(tempfile_path, mode))

    def open(self, mode="rb"):
        if self.file is None:
            self.file = self._download_file(mode)
        elif self.file.closed:
            self.file.open(mode)
        else:
            self.seek(0)
        return self

    def read(self, size=None):
        if not self.closed:
            return self.file.read(size)
        raise ValueError("Cannot read from file.")

    def write(self, chunk):
        if not self.closed:
            return self.file.write(chunk)
        raise ValueError("Cannot write to file.")

    def close(self):
        self.file.close()

    def multiple_chunks(self, chunk_size=None):
        return self.size > (chunk_size or self.DEFAULT_CHUNK_SIZE)

    def chunks(self, chunk_size=None):
        chunk_size = chunk_size or self.DEFAULT_CHUNK_SIZE
        try:
            self.seek(0)
        except (AttributeError, UnsupportedOperation):
            pass

        while True:
            data = self.read(chunk_size)
            if not data:
                break
            yield data


class CloudinaryFileResource(ReadonlyCloudinaryFileProxyMixin, FileResource):
    class Meta(FileResource.Meta):
        abstract = True

    @property
    def name(self) -> str:
        self._require_file()
        return self.get_file().name

    def get_file_folder(self) -> str:
        """
        Возвращает путь к папке, в которую будет сохранен файл.
        Результат вызова используется в параметре `folder` для Cloudinary.
        """
        return ""

    def get_file(self) -> CloudinaryFieldFile:
        raise NotImplementedError

    def get_file_size(self) -> int:
        self._require_file()
        return self.get_file().size

    def get_file_url(self) -> str:
        self._require_file()
        return self.get_file().url

    def file_exists(self) -> bool:
        file = self.get_file()
        if not file:
            return False

        file_field = self.get_file_field()

        try:
            uploader.explicit(
                self.name,
                type=file_field.type,
                resource_type=file_field.resource_type,
            )
        except cloudinary.exceptions.Error:
            return False
        return True

    def get_cloudinary_options(self):
        options = (settings.CLOUDINARY_UPLOADER_OPTIONS or {}).copy()

        file_field = self.get_file_field()
        options.update(file_field.options)

        # owner field`s options should override `CloudinaryField` options
        owner_field = self.get_owner_field()
        if owner_field is not None and hasattr(owner_field, "cloudinary"):
            options.update(owner_field.cloudinary or {})

        # filter keys
        options = {
            key: value
            for key, value in options.items()
            if key in upload_params + uploader.upload_options
        }

        # format folder
        folder = options.pop("folder", None)
        if folder is None:
            folder = self.get_file_folder()
        if isinstance(folder, str):
            options["folder"] = datetime.datetime.now().strftime(folder)

        return options

    def _attach_file(self, file: File, **options):
        cloudinary_options = self.get_cloudinary_options()
        cloudinary_options.update(options.get("cloudinary", {}))

        if file.size >= 100 * 1024 * 1024:
            upload = uploader.upload_large
        else:
            upload = uploader.upload

        file_field = self.get_file_field()
        cloudinary_options.setdefault("type", file_field.type)
        cloudinary_options.setdefault("resource_type", file_field.resource_type)

        try:
            result = upload(
                file,
                **cloudinary_options,
            )
        except cloudinary.exceptions.Error as e:
            if e.args and "Unsupported file type" in e.args[0]:
                raise exceptions.UnsupportedFileError(
                    _("File `%s` is not an image") % file.name
                )
            else:
                raise ValidationError(*e.args)

        # fix difference between `public_id` in response
        # and `public_id` in CloudinaryField
        file_format = result.get("format")
        if file_format:
            public_id = result["public_id"]
        else:
            public_id, file_format = posixpath.splitext(result["public_id"])
            file_format = file_format.lstrip(".")

        resource = CloudinaryResource(
            public_id,
            version=str(result["version"]),
            format=file_format,
            type=result["type"],
            resource_type=result["resource_type"],
            metadata=result,
        )

        self.set_file(resource)

        self.basename = helpers.get_filename(file.name)
        self.extension = file_format or ""
        return result

    def _rename_file(self, new_name: str, **options):
        # TODO: Cloudinary can't copy files. We dont't want to do it manually
        cloudinary_options = self.get_cloudinary_options()
        cloudinary_options.update(options.get("cloudinary", {}))

        file_field = self.get_file_field()
        cloudinary_options.setdefault("type", file_field.type)
        cloudinary_options.setdefault("resource_type", file_field.resource_type)

        old_name = self.name
        folder, basename = posixpath.split(old_name)
        if cloudinary_options["resource_type"] != "raw":
            # video and image have no extension
            base, ext = posixpath.splitext(new_name)
            new_name = base
        new_name = posixpath.join(folder, new_name)

        try:
            result = uploader.rename(
                old_name,
                new_name,
                **cloudinary_options
            )
        except cloudinary.exceptions.Error as e:
            raise ValidationError(*e.args)

        # fix difference between `public_id` in response
        # and `public_id` in CloudinaryField
        file_format = result.get("format")
        if file_format:
            public_id = result["public_id"]
        else:
            public_id, file_format = posixpath.splitext(result["public_id"])
            file_format = file_format.lstrip(".")

        resource = CloudinaryResource(
            public_id,
            version=str(result["version"]),
            format=file_format,
            type=result["type"],
            resource_type=result["resource_type"],
            metadata=result,
        )
        self.set_file(resource)

        self.basename = helpers.get_filename(new_name)
        self.extension = file_format or ""
        return result

    def _delete_file(self, **options):
        cloudinary_options = self.get_cloudinary_options()
        cloudinary_options.update(options.get("cloudinary", {}))

        file = self.get_file()
        if not file:
            return

        file_field = self.get_file_field()
        cloudinary_options.setdefault("type", file_field.type)
        cloudinary_options.setdefault("resource_type", file_field.resource_type)

        try:
            result = uploader.destroy(
                self.name,
                **cloudinary_options
            )
        except cloudinary.exceptions.Error:
            logger.exception(
                "Couldn't delete Cloudinary file: {}".format(self.name)
            )
            return

        status = result.get("result")
        if status == "ok":
            self.set_file(None)
        else:
            logger.warning(
                "Unable to delete Cloudinary file `{}`: {}".format(
                    self.name, status
                )
            )
        return result

    def build_url(self, **options):
        # proxy Cloudinary method
        self._require_file()
        file = self.get_file()
        return file.resource.build_url(**options)



from typing import Optional, Type

from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django.db import models

from ..logging import logger


class BacklinkModelMixin(models.Model):
    """
    Миксина, позволяющая обратиться к полю модели, которое ссылается
    на текущий объект.
    """

    owner_app_label = models.CharField(max_length=100, editable=False)
    owner_model_name = models.CharField(max_length=100, editable=False)
    owner_fieldname = models.CharField(max_length=255, editable=False)

    class Meta:
        abstract = True

    def get_owner_model(self) -> Optional[Type[models.Model]]:
        if not self.owner_app_label or not self.owner_model_name:
            return None

        try:
            return apps.get_model(self.owner_app_label, self.owner_model_name)
        except LookupError:
            logger.debug(
                "Not found model: %s.%s" % (self.owner_app_label, self.owner_model_name)
            )

    def get_owner_field(self) -> Optional[models.Field]:
        owner_model = self.get_owner_model()
        if owner_model is None:
            return None

        if not self.owner_fieldname:
            return None

        try:
            return owner_model._meta.get_field(self.owner_fieldname)
        except FieldDoesNotExist:
            logger.debug(
                "Not found field '%s' in model %s.%s"
                % (self.owner_app_label, self.owner_model_name, self.owner_fieldname)
            )


class FileProxyMixin:
    """
    Проксирование некоторых свойств файла на уровень модели.

    Открытый файл сохраняется в поле `__file`, чтобы предотвратить
    повторное скачивание файла при повторном открытии файла.
    """

    def __init__(self, *args, **kwargs):
        self.__file = None
        super().__init__(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def _get_file(self):
        return getattr(self, "_FileProxyMixin__file", None)

    def _open_file(self, mode):
        self.__file = self.get_file().open(mode)  # noqa: F821

    def _close_file(self):
        file = self._get_file()
        if file is not None:
            file.close()
        self.__file = None

    @property
    def closed(self):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        return not file or file.closed

    def open(self, mode="rb"):
        self._require_file()  # noqa: F821

        file = self._get_file()
        if file is None:
            self._open_file(mode)
        elif file.seekable():
            file.seek(0)
        else:
            # current file is not seekable - reopen it
            file.close()
            self._open_file(mode)
        return self
    open.alters_data = True  # noqa: F821

    def read(self, size=None):
        self._require_file()  # noqa: F821

        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821

        return file.read(size)  # noqa: F821

    def close(self):
        self._close_file()

    def readable(self):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        if hasattr(file, "readable"):
            return file.readable()
        return True

    def writable(self):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        if hasattr(file, "writable"):
            return file.writable()
        return "w" in getattr(file, "mode", "")

    def seekable(self):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        if hasattr(file, "seekable"):
            return file.seekable()
        return True

    def seek(self, *args, **kwargs):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        return file.seek(*args, **kwargs)

    def tell(self):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        return file.tell()

    def multiple_chunks(self, chunk_size=None):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        return file.multiple_chunks(chunk_size)

    def chunks(self, chunk_size=None):
        file = self._get_file()
        if file is None:
            file = self.get_file()  # noqa: F821
        return file.chunks(chunk_size)


class FileFieldProxyMixin:
    """
    Проксирование некоторых свойств файла (только для чтения) на уровень модели
    """

    @property
    def path(self):
        self._require_file()  # noqa: F821
        return self.get_file().path  # noqa: F821

    @property
    def url(self):
        self._require_file()  # noqa: F821
        return self.get_file().url  # noqa: F821



from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.core.files.uploadedfile import UploadedFile
from django.db import transaction
from django.utils.decorators import method_decorator
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt

from .. import exceptions, signals
from ..helpers import run_validators
from ..logging import logger
from ..models.collection import CollectionBase, CollectionItemBase
from . import helpers
from .base import ActionView, ChangeFileViewBase, DeleteFileViewBase, UploadFileViewBase


class CreateCollectionView(ActionView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        if not request.user.has_perm("paper_uploads.upload"):
            return self.error_response(_("Access denied"))
        return self.perform_action(request, *args, **kwargs)

    def handle(self, request, *args, **kwargs):
        content_type_id = request.POST.get("paperCollectionContentType")
        collection_cls = helpers.get_model_class(content_type_id, CollectionBase)
        collection = collection_cls.objects.create(
            owner_app_label=request.POST.get("paperOwnerAppLabel"),
            owner_model_name=request.POST.get("paperOwnerModelName"),
            owner_fieldname=request.POST.get("paperOwnerFieldName"),
        )
        return self.success_response({
            "collection_id": collection.pk
        })


class DeleteCollectionView(ActionView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        if not request.user.has_perm("paper_uploads.delete"):
            return self.error_response(_("Access denied"))
        return self.perform_action(request, *args, **kwargs)

    def handle(self, request, *args, **kwargs):
        content_type_id = request.POST.get("paperCollectionContentType")
        collection_cls = helpers.get_model_class(content_type_id, CollectionBase)
        collection_id = request.POST.get("pk")

        try:
            instance = helpers.get_instance(collection_cls, collection_id)
        except exceptions.InvalidObjectId:
            logger.exception("Error")
            return self.error_response(_("Invalid ID"))
        except ObjectDoesNotExist:
            logger.exception("Error")
            return self.error_response(_("Object not found"))
        except MultipleObjectsReturned:
            logger.exception("Error")
            return self.error_response(_("Multiple objects returned"))

        instance.delete()
        return self.success_response()


class UploadFileView(UploadFileViewBase):
    def handle(self, request, file: UploadedFile):
        content_type_id = request.POST.get("paperCollectionContentType")
        collection_cls = helpers.get_model_class(content_type_id, CollectionBase)
        collection_id = request.POST.get("collectionId")

        try:
            collection = helpers.get_instance(collection_cls, collection_id)
        except exceptions.InvalidObjectId:
            logger.exception("Error")
            return self.error_response(_("Invalid collection ID"))
        except ObjectDoesNotExist:
            logger.exception("Error")
            return self.error_response(_("Collection not found"))
        except MultipleObjectsReturned:
            logger.exception("Error")
            return self.error_response(_("Multiple objects returned"))

        try:
            order = max(0, int(request.POST.get("order")))
        except (TypeError, ValueError):
            order = 0

        # перебираем все подходящие классы элементов пока
        # не найдем тот, который будет успешно создан
        for item_type in collection.detect_item_type(file):  # noqa: F821
            item_type_field = collection_cls.item_types[item_type]
            instance = item_type_field.model(
                collection_content_type_id=content_type_id,
                collection_id=collection.pk,
                item_type=item_type,
                size=file.size,
                order=order
            )

            try:
                instance.attach_file(file)
            except exceptions.UnsupportedFileError:
                continue

            try:
                instance.full_clean()
                run_validators(file, item_type_field.validators)
            except Exception:
                instance.delete_file()
                raise

            break
        else:
            return self.error_response(_("Unsupported file: %s") % file.name)

        instance.save()
        return self.success_response({
            **instance.as_dict(),
        })


class DeleteFileView(DeleteFileViewBase):
    def handle(self, request):
        content_type_id = request.POST.get("paperCollectionContentType")
        collection_cls = helpers.get_model_class(content_type_id, CollectionBase)

        item_type = request.POST.get("itemType")
        for name, field in collection_cls.item_types.items():
            if item_type == name:
                model_class = field.model
                break
        else:
            return self.error_response(_("Invalid itemType"))

        item_id = request.POST.get("itemId")

        try:
            item = helpers.get_instance(model_class, item_id)
        except exceptions.InvalidObjectId:
            logger.exception("Error")
            return self.error_response(_("Invalid ID"))
        except ObjectDoesNotExist:
            # silently skip
            return self.success_response()
        except MultipleObjectsReturned:
            logger.exception("Error")
            return self.error_response(_("Multiple objects returned"))

        item.delete()
        return self.success_response()


class ChangeFileView(ChangeFileViewBase):
    template_name = "paper_uploads/dialogs/collection.html"

    def get_form_class(self):
        return import_string(self.instance.change_form_class)

    def get_instance(self, request, *args, **kwargs):
        content_type_id = self.request.GET.get("paperCollectionContentType")
        collection_cls = helpers.get_model_class(content_type_id, CollectionBase)

        item_type = self.request.GET.get("itemType")
        for name, field in collection_cls.item_types.items():
            if item_type == name:
                model_class = field.model
                break
        else:
            raise exceptions.AjaxFormError(_("Invalid itemType"))

        item_id = self.request.GET.get("itemId")

        try:
            return helpers.get_instance(model_class, item_id)
        except exceptions.InvalidObjectId:
            raise exceptions.AjaxFormError(_("Invalid ID"))
        except ObjectDoesNotExist:
            raise exceptions.AjaxFormError(_("Object not found"))
        except MultipleObjectsReturned:
            raise exceptions.AjaxFormError(_("Multiple objects returned"))


class SortItemsView(ActionView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        if not request.user.has_perm("paper_uploads.change"):
            return self.error_response(_("Access denied"))

        return self.perform_action(request, *args, **kwargs)

    def handle(self, request, *args, **kwargs):
        content_type_id = request.POST.get("paperCollectionContentType")
        collection_cls = helpers.get_model_class(content_type_id, CollectionBase)
        collection_id = request.POST.get("collectionId")

        try:
            instance = helpers.get_instance(collection_cls, collection_id)
        except exceptions.InvalidObjectId:
            logger.exception("Error")
            return self.error_response(_("Invalid ID"))
        except ObjectDoesNotExist:
            logger.exception("Error")
            return self.error_response(_("Object not found"))
        except MultipleObjectsReturned:
            logger.exception("Error")
            return self.error_response(_("Multiple objects returned"))

        order_string = request.POST.get("orderList", "")
        try:
            item_ids = (int(pk) for pk in order_string.split(","))
        except ValueError:
            logger.exception("Error")
            return self.error_response(_("Invalid order value"))

        with transaction.atomic():
            for index, item_id in enumerate(item_ids):
                if item_id in set(instance.items.values_list("pk", flat=True)):  # noqa: F821
                    CollectionItemBase.objects.filter(pk=item_id).update(order=index)
                else:
                    CollectionItemBase.objects.filter(pk=item_id).update(
                        order=2 ** 32 - 1
                    )

        signals.collection_reordered.send(
            sender=collection_cls,
            instance=instance
        )
        return self.success_response({
            "orderMap": dict(instance.items.values_list("pk", "order"))  # noqa: F821
        })



from typing import Any, Dict, Optional

import magic
from cloudinary.models import CloudinaryField
from django.core.files import File
from django.db import models
from django.utils.translation import gettext_lazy as _

from ...conf import settings
from ...models.base import ImageFileResourceMixin
from ...models.collection import Collection, CollectionItemBase, FilePreviewMixin
from ...models.fields import CollectionItem
from .base import CloudinaryFieldFile, CloudinaryFileResource

__all__ = [
    "CloudinaryFileItem",
    "CloudinaryMediaItem",
    "CloudinaryImageItem",
    "CloudinaryCollection",
    "CloudinaryImageCollection",
]


class CollectionCloudinaryFileItemBase(CollectionItemBase, CloudinaryFileResource):
    """
    Базовый класс элемента галереи, содержащего файл.
    """

    class Meta:
        abstract = True

    @classmethod
    def file_supported(cls, file: File) -> bool:
        """
        Проверка возможности представления загруженного файла
        текущим классом элемента в коллекции.
        """
        raise NotImplementedError


class CloudinaryFileItem(FilePreviewMixin, CollectionCloudinaryFileItemBase):
    change_form_class = "paper_uploads.forms.dialogs.collection.FileItemDialog"
    template_name = "paper_uploads/items/file.html"
    preview_template_name = "paper_uploads/items/preview/file.html"

    file = CloudinaryField(
        _("file"),
        type=settings.CLOUDINARY_TYPE,
        resource_type="raw",
    )
    display_name = models.CharField(_("display name"), max_length=255, blank=True)

    class Meta(CollectionCloudinaryFileItemBase.Meta):
        verbose_name = _("File item")
        verbose_name_plural = _("File items")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.basename
        super().save(*args, **kwargs)

    def get_file_folder(self) -> str:
        return settings.COLLECTION_FILES_UPLOAD_TO

    def get_file(self) -> Optional[CloudinaryFieldFile]:
        if not self.file:
            return None
        return CloudinaryFieldFile(self.file, checksum=self.checksum)

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> CloudinaryField:
        return self._meta.get_field("file")

    @classmethod
    def file_supported(cls, file: File) -> bool:
        return True


class CloudinaryMediaItem(FilePreviewMixin, CollectionCloudinaryFileItemBase):
    change_form_class = "paper_uploads.forms.dialogs.collection.FileItemDialog"
    template_name = "paper_uploads/items/file.html"
    preview_template_name = "paper_uploads/items/preview/file.html"

    file = CloudinaryField(
        _("file"),
        type=settings.CLOUDINARY_TYPE,
        resource_type="video",
    )
    display_name = models.CharField(_("display name"), max_length=255, blank=True)

    class Meta(CollectionCloudinaryFileItemBase.Meta):
        verbose_name = _("Media item")
        verbose_name_plural = _("Media items")

    def save(self, *args, **kwargs):
        if not self.pk and not self.display_name:
            self.display_name = self.basename
        super().save(*args, **kwargs)

    def get_file_folder(self) -> str:
        return settings.COLLECTION_FILES_UPLOAD_TO

    def get_file(self) -> Optional[CloudinaryFieldFile]:
        if not self.file:
            return None
        return CloudinaryFieldFile(self.file, checksum=self.checksum)

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> CloudinaryField:
        return self._meta.get_field("file")

    @classmethod
    def file_supported(cls, file: File) -> bool:
        mimetype = magic.from_buffer(file.read(1024), mime=True)
        file.seek(0)  # correct file position after mimetype detection
        basetype, subtype = mimetype.split("/", 1)
        return basetype in {"video", "audio"}


class CloudinaryImageItem(ImageFileResourceMixin, CollectionCloudinaryFileItemBase):
    PREVIEW_VARIATIONS = settings.COLLECTION_IMAGE_ITEM_PREVIEW_VARIATIONS
    change_form_class = "paper_uploads.forms.dialogs.collection.ImageItemDialog"
    template_name = "paper_uploads_cloudinary/items/image.html"
    preview_template_name = "paper_uploads_cloudinary/items/preview/image.html"

    file = CloudinaryField(
        _("file"),
        type=settings.CLOUDINARY_TYPE,
        resource_type="image",
    )

    class Meta(CollectionCloudinaryFileItemBase.Meta):
        verbose_name = _("Image item")
        verbose_name_plural = _("Image items")

    def get_file_folder(self) -> str:
        return settings.COLLECTION_IMAGES_UPLOAD_TO

    def get_file(self) -> Optional[CloudinaryFieldFile]:
        if not self.file:
            return None
        return CloudinaryFieldFile(self.file, checksum=self.checksum)

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> CloudinaryField:
        return self._meta.get_field("file")

    @classmethod
    def file_supported(cls, file: File) -> bool:
        mimetype = magic.from_buffer(file.read(1024), mime=True)
        file.seek(0)  # correct file position after mimetype detection
        basetype, subtype = mimetype.split("/", 1)
        return basetype == "image"


# ==============================================================================


class CloudinaryImageCollection(Collection):
    image = CollectionItem(CloudinaryImageItem)

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "strictImageValidation": True,
            "acceptFiles": [
                "image/bmp",
                "image/gif",
                "image/jpeg",
                "image/png",
                # "image/svg+xml",
                "image/tiff",
                "image/webp",
            ],
        }



from typing import Any, Dict

from django.db.models.fields.files import FieldFile
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..storage import upload_storage
from ..utils import filesizeformat
from ..variations import PaperVariation
from .base import FileFieldResource, VersatileImageResourceMixin
from .fields import VariationalFileField
from .utils import generate_filename


class UploadedImage(VersatileImageResourceMixin, FileFieldResource):
    file = VariationalFileField(
        _("file"),
        max_length=255,
        storage=upload_storage,
        upload_to=generate_filename,
    )

    class Meta(FileFieldResource.Meta):
        verbose_name = _("image")
        verbose_name_plural = _("images")

    def get_file_folder(self) -> str:
        return settings.IMAGES_UPLOAD_TO

    def get_file(self) -> FieldFile:
        return self.file

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> VariationalFileField:
        return self._meta.get_field("file")

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "file_info": "({ext}, {width}x{height}, {size})".format(
                ext=self.extension,
                width=self.width,
                height=self.height,
                size=filesizeformat(self.size),
            ),
        }

    def get_variations(self) -> Dict[str, PaperVariation]:
        if not hasattr(self, "_variations_cache"):
            owner_field = self.get_owner_field()
            if owner_field is not None:
                self._variations_cache = getattr(owner_field, "variations", {}).copy()
            else:
                return {}
        return self._variations_cache

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "strictImageValidation": True,
            "acceptFiles": [
                "image/bmp",
                "image/gif",
                "image/jpeg",
                "image/png",
                # "image/svg+xml",
                "image/tiff",
                "image/webp",
            ],
        }



from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.core.files.uploadedfile import UploadedFile
from django.utils.translation import gettext_lazy as _

from .. import exceptions
from ..forms.dialogs.image import UploadedImageDialog
from ..helpers import run_validators
from ..logging import logger
from ..models.base import FileResource
from . import helpers
from .base import ChangeFileViewBase, DeleteFileViewBase, UploadFileViewBase


class UploadFileView(UploadFileViewBase):
    def handle(self, request, file: UploadedFile):
        content_type_id = request.POST.get("paperContentType")
        model_class = helpers.get_model_class(content_type_id, FileResource)
        instance = model_class(
            owner_app_label=request.POST.get("paperOwnerAppLabel"),
            owner_model_name=request.POST.get("paperOwnerModelName"),
            owner_fieldname=request.POST.get("paperOwnerFieldName"),
        )

        try:
            instance.attach_file(file)
        except exceptions.UnsupportedFileError as e:
            return self.error_response(e.message)

        try:
            instance.full_clean()
            owner_field = instance.get_owner_field()
            if owner_field is not None:
                run_validators(file, owner_field.validators)
        except Exception:
            instance.delete_file()
            raise

        instance.save()
        return self.success_response(instance.as_dict())


class DeleteFileView(DeleteFileViewBase):
    def handle(self, request):
        content_type_id = request.POST.get("paperContentType")
        model_class = helpers.get_model_class(content_type_id, FileResource)
        pk = request.POST.get("pk")

        try:
            instance = helpers.get_instance(model_class, pk)
        except exceptions.InvalidObjectId:
            logger.exception("Error")
            return self.error_response(_("Invalid ID"))
        except ObjectDoesNotExist:
            logger.exception("Error")
            return self.error_response(_("Object not found"))
        except MultipleObjectsReturned:
            logger.exception("Error")
            return self.error_response(_("Multiple objects returned"))

        instance.delete()
        return self.success_response()


class ChangeFileView(ChangeFileViewBase):
    form_class = UploadedImageDialog
    template_name = "paper_uploads/dialogs/image.html"

    def get_instance(self, request, *args, **kwargs):
        content_type_id = self.request.GET.get("paperContentType")
        model_class = helpers.get_model_class(content_type_id, FileResource)
        pk = self.request.GET.get("pk")

        try:
            return helpers.get_instance(model_class, pk)
        except exceptions.InvalidObjectId:
            raise exceptions.AjaxFormError(_("Invalid ID"))
        except ObjectDoesNotExist:
            raise exceptions.AjaxFormError(_("Object not found"))
        except MultipleObjectsReturned:
            raise exceptions.AjaxFormError(_("Multiple objects returned"))



from django.apps import AppConfig, apps
from django.db.models.signals import pre_migrate
from django.utils.translation import gettext_lazy as _


class Config(AppConfig):
    name = "paper_uploads"
    verbose_name = _("Uploaded Files")

    def ready(self):
        from django.contrib.admin.sites import site
        from .admin import CollectionAdminBase
        from .models.collection import CollectionBase
        from .signals import handlers

        # Переименование поля файла или модели-владельца файла,
        # изменит также и соответствующие значения, хранящиеся в БД
        pre_migrate.connect(handlers.inject_rename_filefield_operations, sender=self)

        # автоматическая регистрация моделей галерей в админке
        for model in apps.get_models():
            if issubclass(model, CollectionBase) and not site.is_registered(model):
                site.register(model, CollectionAdminBase)



import os
import time
from typing import Any, Dict, Iterable, Iterator, List, Set

from django.apps import apps
from django.core import exceptions
from django.core.exceptions import ObjectDoesNotExist
from django.db import DEFAULT_DB_ALIAS

from .conf import settings
from .logging import logger
from .typing import FileLike, VariationConfig
from .utils import lowercased_dict_keys
from .variations import PaperVariation

# Перечень допустимых версий вариаций
ALLOWED_VERSIONS = {"webp", "2x", "3x", "4x"}

# Максимальное количество попыток чтения из БД при получении
# экземпляра модели. Сделано потому, что каждый запрос оборачивается
# в транзакцию. Из-за этого вызов `recut()` (даже после метода `save()`)
# не гарантирует, что данные уже попали в БД.
# TODO: поискать альтернативные решения
MAX_DB_READ_ATTEMPTS = 3


def get_filename(filename: str) -> str:
    basename = os.path.basename(filename)
    name, _ = os.path.splitext(basename)
    return name


def get_extension(filename: str) -> str:
    basename = os.path.basename(filename)
    _, extension = os.path.splitext(basename)
    return extension.lstrip(".")


def generate_scaled_versions(
    name: str, config: VariationConfig, scale_factor: int = 1, webp: bool = False
) -> Iterator[PaperVariation]:
    """
    Геренирует Retina-версию вариации с опциональной WebP-версией того же размера.
    """
    scaled_size = tuple(x * scale_factor for x in config.get("size", (0, 0)))

    if scale_factor == 1:
        variation_name = name
    else:
        variation_name = "{}_{}x".format(name, scale_factor)

    variation_config = dict(
        config,
        name=variation_name,
        size=scaled_size,
        max_width=scale_factor * config.get("max_width", 0),
        max_height=scale_factor * config.get("max_height", 0),
    )

    yield PaperVariation(**variation_config)

    if webp:
        if scale_factor == 1:
            variation_name = "{}_webp".format(name)
        else:
            variation_name = "{}_webp_{}x".format(name, scale_factor)

        variation_config = dict(
            config,
            name=variation_name,
            size=scaled_size,
            max_width=scale_factor * config.get("max_width", 0),
            max_height=scale_factor * config.get("max_height", 0),
            format="webp"
        )

        yield PaperVariation(**variation_config)


def generate_variation_versions(
    name: str, config: VariationConfig, versions: Set[str],
) -> Iterator[PaperVariation]:
    """
    Геренирует все указанные версии вариации по словарю конфигурации
    """
    webp = "webp" in versions and config.get("format", "").upper() != "WEBP"
    yield from generate_scaled_versions(name, config, scale_factor=1, webp=webp)
    if "2x" in versions:
        yield from generate_scaled_versions(name, config, scale_factor=2, webp=webp)
    if "3x" in versions:
        yield from generate_scaled_versions(name, config, scale_factor=3, webp=webp)
    if "4x" in versions:
        yield from generate_scaled_versions(name, config, scale_factor=4, webp=webp)


def build_variations(options: Dict[str, VariationConfig]) -> Dict[str, PaperVariation]:
    """
    Создание объектов вариаций из словаря конфигураций.
    """
    variations = {}
    for name, config in options.items():
        new_config = lowercased_dict_keys(settings.VARIATION_DEFAULTS or {})
        new_config.update(config)

        versions = set(v.lower() for v in new_config.get("versions", ()))
        unknown_versions = versions.difference(ALLOWED_VERSIONS)
        if unknown_versions:
            raise ValueError(
                "unknown variation versions: {}".format(", ".join(unknown_versions))
            )

        all_versions = generate_variation_versions(name, new_config, versions)
        for variation in all_versions:
            if name == variation.name:
                # явно заданная вариация переопредеяет любую неявную
                variations[name] = variation
            else:
                variations.setdefault(variation.name, variation)

    return variations


def get_instance(
    app_label: str, model_name: str, object_id: int, using: str = DEFAULT_DB_ALIAS
):
    """
    Получение экземпляра модели по названию приложения, модели и ID.
    """
    model_class = apps.get_model(app_label, model_name)
    attempts = 1
    while True:
        try:
            return model_class._base_manager.using(using).get(pk=object_id)
        except ObjectDoesNotExist:
            # delay recheck if transaction not committed yet
            attempts += 1
            if attempts > MAX_DB_READ_ATTEMPTS:
                logger.exception("Instance #%s not found" % object_id)
                raise
            else:
                time.sleep(1)


def run_validators(value: FileLike, validators: Iterable[Any]):
    """
    Based on `django.forms.fields.run_validators` method.
    """
    errors = []  # type: List[Any]
    for v in validators:
        try:
            v(value)
        except exceptions.ValidationError as e:
            errors.extend(e.error_list)

    if errors:
        raise exceptions.ValidationError(errors)


def _get_item_types(cls):
    """
    Получение явно объявленных (не унаследованных) типов элементов коллекции
    из её класса.
    """
    attname = "_{}__item_types".format(cls.__name__)
    return getattr(cls, attname, None)


def _set_item_types(cls, value):
    attname = "_{}__item_types".format(cls.__name__)
    setattr(cls, attname, value)



from typing import Any, Dict, Optional

from cloudinary.models import CloudinaryField
from django.utils.translation import gettext_lazy as _

from ...conf import settings
from ...models.base import ImageFileResourceMixin
from ...utils import filesizeformat
from .base import CloudinaryFieldFile, CloudinaryFileResource


class CloudinaryImage(ImageFileResourceMixin, CloudinaryFileResource):
    file = CloudinaryField(
        _("file"),
        type=settings.CLOUDINARY_TYPE,
        resource_type="image",
    )

    class Meta(CloudinaryFileResource.Meta):
        verbose_name = _("image")
        verbose_name_plural = _("images")

    def get_file_folder(self) -> str:
        return settings.IMAGES_UPLOAD_TO

    def get_file(self) -> Optional[CloudinaryFieldFile]:
        if not self.file:
            return None
        return CloudinaryFieldFile(self.file, checksum=self.checksum)

    def set_file(self, value):
        self.file = value

    def get_file_field(self) -> CloudinaryField:
        return self._meta.get_field("file")

    def as_dict(self) -> Dict[str, Any]:
        return {
            **super().as_dict(),
            "file_info": "({ext}, {width}x{height}, {size})".format(
                ext=self.extension,
                width=self.width,
                height=self.height,
                size=filesizeformat(self.size),
            ),
        }

    @classmethod
    def get_configuration(cls) -> Dict[str, Any]:
        return {
            "strictImageValidation": True,
            "acceptFiles": [
                "image/bmp",
                "image/gif",
                "image/jpeg",
                "image/png",
                # "image/svg+xml",
                "image/tiff",
                "image/webp",
            ],
        }



from ...models.file import UploadedFile
from .base import UploadedFileBaseForm


class UploadedFileDialog(UploadedFileBaseForm):
    class Meta:
        model = UploadedFile
        fields = ("new_name", "display_name")



from typing import Any, Dict

from django.core import checks
from django.db import models
from django.db.models.signals import post_delete

from ... import validators


class ResourceFieldBase(models.OneToOneField):
    """
    Базовый класс для ссылок на ресурсы.
    """

    def __init__(self, verbose_name=None, **kwargs):
        kwargs.setdefault("null", True)
        kwargs.setdefault("related_name", "+")
        kwargs.setdefault("on_delete", models.SET_NULL)
        kwargs["verbose_name"] = verbose_name
        super().__init__(**kwargs)

    def check(self, **kwargs):
        return [*super().check(**kwargs), *self._check_relation()]

    def _check_relation(self):
        from ...models.base import Resource

        return self._check_relation_class(
            Resource,
            "Field defines a relation with model '%s', "
            "which is not subclass of Resource model",
        )

    def _check_relation_class(self, base, error_message):
        rel_is_string = isinstance(self.remote_field.model, str)
        if rel_is_string:
            return []

        model_name = (
            self.remote_field.model
            if rel_is_string
            else self.remote_field.model._meta.object_name
        )

        if not issubclass(self.remote_field.model, base):
            return [checks.Error(error_message % model_name, obj=self)]
        return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if "null" in kwargs and self.null:
            del kwargs["null"]
        if "related_name" in kwargs and kwargs["related_name"] == "+":
            del kwargs["related_name"]
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "owner_app_label": self.opts.app_label.lower(),
                "owner_model_name": self.opts.model_name.lower(),
                "owner_fieldname": self.name,
                **kwargs,
            }
        )

    def run_validators(self, value):
        """
        Валидаторы мы хотим запускать по отношению в экземпляру File,
        а не к ID экземпляа модели.
        """
        pass

    def contribute_to_class(self, cls, *args, **kwargs):
        super().contribute_to_class(cls, *args, **kwargs)
        if not cls._meta.abstract:
            post_delete.connect(self.post_delete_callback, sender=cls)

    def post_delete_callback(self, **kwargs):
        """
        При удалении модели-владельца поля, удаляем и связанный
        в ним экземпляр ресурса.
        """
        resource = getattr(kwargs["instance"], self.name)
        if resource:
            resource.delete()


class FileResourceFieldBase(ResourceFieldBase):
    """
    Базовый класс для полей, которые загружают файлы.
    """

    def formfield(self, **kwargs):
        return super().formfield(**{"configuration": self.get_configuration(), **kwargs})

    def get_configuration(self) -> Dict[str, Any]:
        """
        Превращает Django-валидаторы в словарь конфигурации,
        который может использоваться для вывода или проверки
        на стороне клиента.
        """
        config = {}
        for v in self.validators:
            if isinstance(v, validators.MimeTypeValidator):
                config["acceptFiles"] = v.allowed
            elif isinstance(v, validators.ExtensionValidator):
                config["allowedExtensions"] = v.allowed
            elif isinstance(v, validators.SizeValidator):
                config["sizeLimit"] = v.limit_value
            elif isinstance(v, validators.ImageMinSizeValidator):
                config["minImageWidth"] = v.width_limit
                config["minImageHeight"] = v.height_limit
            elif isinstance(v, validators.ImageMaxSizeValidator):
                config["maxImageWidth"] = v.width_limit
                config["maxImageHeight"] = v.height_limit
        return config
